#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "rbmp.h"

//#define BMP_DEBUG_LOW
#define BMP_DEBUG_MED
#define BMP_DEBUG_HIGH

unsigned char *imagebuffer;
int imagewidth, imageheight;
int imagepos;

int currentoffsetx, currentoffsety;
int currentwidth, currentheight;
int globalmapinfo, globalmappresent, globalbpp, globalcpp;
int localmapinfo, localmappresent, localcpp, localimagetype;

unsigned char *bmpcurbyte;
unsigned char *bmpbuffer;
unsigned char *bmpmaxbyte;
unsigned char background;
int bmplength;
long bmpbitsread = 0;
int bmpbytecounter;
int bmpblocksizebyte;
int bmplastcode;
int bmplastcodenum;

unsigned char redcolormap[256];
unsigned char greencolormap[256];
unsigned char bluecolormap[256];
unsigned char alphacolormap[256];
int numcolors;

/*
int main(int argc, char **argv){
	int width;
	int height;
	int comps;
	unsigned char *buffer;
	
	return (readBMP(argv[1], &buffer, &width, &height, &comps));	
}
*/

int readBMP(char *filename, unsigned char **buffer, int *width, int *height, int *components){
	int filesize, imageoffset, imagesize, headersize;
	unsigned short bpp;
	unsigned short numplanes;
	int resx, resy;
	int numcolors, impcolors;
	int compression;	
	int numpixels, linesize, fillerbytes;
	char imageident[3];

	int i, j, k, l;
	int r, g, b, a;

	imagebuffer = NULL;

	if (openBMPfile(filename) == 0){
		printf("Error opening BMP file %s\n", filename);
		return(0);
	}

	getBMPbuffer(imageident, 2);
	imageident[2] = 0;
	filesize = getBMPint();
	getBMPshort();
	getBMPshort();
	imageoffset = getBMPint();

	//#ifdef BMP_DEBUG_LOW
	printf("Magic ID: %s\n", imageident);
	printf("Filesize: %d\n", filesize);
	printf("Image Offset: %d\n", imageoffset);
	//#endif

	/* now get the image info header */
	headersize = getBMPint();
	imagewidth = getBMPint();
	imageheight = getBMPint();
	numplanes = getBMPshort();
	bpp = getBMPshort();
	compression = getBMPint();
	imagesize = getBMPint();
	resx = getBMPint();
	resy = getBMPint();
	numcolors = getBMPint();
	impcolors = getBMPint();

	//#ifdef BMP_DEBUG_LOW
	printf("Info Header Size: %d\n", headersize);
	printf("Image Size: (%d,%d) %d\n", imagewidth, imageheight, imagesize);
	printf("Planes: %d\n", numplanes);
	printf("Bits per Pixel: %d\n", bpp);
	printf("Compression: %d (0 = none, 1 - 8 bit RLE, 2 - 4 bit RLE, 3 - RGB bitmap with mask)\n", compression);
	printf("Device Resolution: %dx%d\n", resx, resy);
	printf("Colors: %d Important %d\n", numcolors, impcolors);
	//#endif


	/* allocate memory for the image */

	imagebuffer = calloc(imagewidth * imageheight * 3, 1);
	if (imagebuffer == NULL){
		printf("Couldn't allocate BMP image buffer memory of size %d\n", currentwidth * currentheight * 3);
		return(0);
	}


	/* get the color palette if it exists */
	/* if 1 bpp then only 2 color entries */

	if (bpp == 1) numcolors = 2;
	else if (bpp == 4) numcolors = 16;
	else if (bpp == 8) numcolors = 256;

	for (i = 0; i < numcolors; i++){
		bluecolormap[i] = getBMPbyte();
		greencolormap[i] = getBMPbyte();
		redcolormap[i] = getBMPbyte();
		alphacolormap[i] = getBMPbyte();
	}
	
	gotoBMPbyte(imageoffset);

	if (compression == 0){
		if (bpp == 1){
			linesize = imagewidth * 3;
			fillerbytes = (imagewidth / 8) % 4;  
			k = (imageheight - 1) * linesize;

			for (j = 0; j < imageheight; j++){
				l = k;
				for (i = 0; i <imagewidth; i++){
					a = getBMPbit();
					imagebuffer[l] = redcolormap[a];
					l++;
					imagebuffer[l] = greencolormap[a];
					l++;
					imagebuffer[l] = bluecolormap[a];
					l++;
				}
				forwardBMPbytes(fillerbytes);
				k -= linesize; 
			}
		}
		if (bpp == 4){
			linesize = imagewidth * 3;
			fillerbytes = (imagewidth / 2) % 4;  
			k = (imageheight - 1) * linesize;

			for (j = 0; j < imageheight; j++){
				l = k;
				for (i = 0; i <imagewidth; i++){
					a = getBMPnibble();
					imagebuffer[l] = redcolormap[a];
					l++;
					imagebuffer[l] = greencolormap[a];
					l++;
					imagebuffer[l] = bluecolormap[a];
					l++;
				}
				forwardBMPbytes(fillerbytes);
				k -= linesize; 
			}
		}
		if (bpp == 8){
			linesize = imagewidth * 3;
			fillerbytes = imagewidth % 4;  
			k = (imageheight - 1) * linesize;

			for (j = 0; j < imageheight; j++){
				l = k;
				for (i = 0; i <imagewidth; i++){
					a = getBMPbyte();
					imagebuffer[l] = redcolormap[a];
					l++;
					imagebuffer[l] = greencolormap[a];
					l++;
					imagebuffer[l] = bluecolormap[a];
					l++;
				}
				forwardBMPbytes(fillerbytes);
				k -= linesize; 
			}
		}
		if (bpp == 24){
			linesize = imagewidth * 3;
			fillerbytes = linesize % 4;  
			k = (imageheight - 1) * linesize;

			for (j = 0; j < imageheight; j++){
				l = k;
				for (i = 0; i <imagewidth; i++){
					b = getBMPbyte();
					g = getBMPbyte();
					r = getBMPbyte();
					imagebuffer[l] = r;
					l++;
					imagebuffer[l] = g;
					l++;
					imagebuffer[l] = b;
					l++;
				}
				forwardBMPbytes(fillerbytes);
				k -= linesize; 
			}
		}
		if (bpp == 32){
			linesize = imagewidth * 3;
			k = (imageheight - 1) * linesize;

			for (j = 0; j < imageheight; j++){
				l = k;
				for (i = 0; i <imagewidth; i++){
					b = getBMPbyte();
					g = getBMPbyte();
					r = getBMPbyte();
					a = getBMPbyte();
					imagebuffer[l] = r;
					l++;
					imagebuffer[l] = g;
					l++;
					imagebuffer[l] = b;
					l++;
				}
				k -= linesize; 
			}
		}
	}

	/* if we're done set the imagesize and buffer values */

	*width = imagewidth;
	*height = imageheight;
	*buffer = imagebuffer;
	*components = 3;
	return(1);
}

int openBMPfile(char *filename){
	FILE *bmpfp;
	unsigned char *buffer;
	int size, read;

	bmpfp = fopen(filename, "rb");
	if (bmpfp == NULL){
		//printf("Error opening BMP file %s\n", filename);
		return(0);
	}

	size = 0;
	if (fseek(bmpfp, 0, SEEK_END) == 0){
		size = ftell(bmpfp);
	}	
	
	if (size == 0) return (0);

	buffer = calloc(size, 1);
	fseek(bmpfp, 0, SEEK_SET);
	read = fread(buffer, 1, size, bmpfp);
	if (read != size){
		printf("Error reading BMP file %s, size %d bytes, read %d bytes\n", filename, size, read); 
		return(0);
	}

	bmpbuffer = buffer;
	bmpcurbyte = buffer; 
	bmplength = size;
	bmpmaxbyte = bmpbuffer + size;

	fclose(bmpfp);
	return(1);
}

int getBMPint(){
	int value;
	int temp;

	if (bmpcurbyte < bmpmaxbyte){ 
		value = (int)bmpcurbyte[0];
		bmpcurbyte++;
		value |= (int)bmpcurbyte[0] << 8;
		bmpcurbyte++;
		value |= (int)bmpcurbyte[0] << 16;
		bmpcurbyte++;
		value |= (int)bmpcurbyte[0] << 24;
		bmpcurbyte++;
		return(value);
	}
	else{
		printf("Fetching out of bounds BMP int %ld/%ld\n", (long)bmpcurbyte, (long)bmpmaxbyte);
		return(-1); 
	}
}

int getBMPshort(){
	int value;
	int temp;

	if (bmpcurbyte < bmpmaxbyte){ 
		value = (int)bmpcurbyte[0];
		bmpcurbyte++;
		value |= (int)bmpcurbyte[0] << 8;
		bmpcurbyte++;
		return(value);
	}
	else{
		printf("Fetching out of bounds BMP int %ld/%ld\n", (long)bmpcurbyte, (long)bmpmaxbyte);
		return(-1); 
	}
}

int getBMPbyte(){
	int value;

	if (bmpcurbyte < bmpmaxbyte){ 
		value = (int)bmpcurbyte[0];
		bmpcurbyte++;
		return(value);
	}
	else{
		printf("Fetching out of bounds BMP byte %ld/%ld\n", (long)bmpcurbyte, (long)bmpmaxbyte);
		return(-1);
	}
}

int getBMPnibble(){
	static unsigned char nibble = 0;
	static int bitbuffer;
	static int value;

	if (bmpcurbyte < bmpmaxbyte){
		if (nibble == 0){
			bitbuffer = (int)bmpcurbyte[0];
			bmpcurbyte++;
			value = bitbuffer >> 4;
			nibble = 1;
		}
		else{
			value = bitbuffer & 0x0f;
			nibble = 0;
		}
		return(value);
	}
	else{
		printf("Fetching out of bounds BMP nibble %ld/%ld\n", (long)bmpcurbyte, (long)bmpmaxbyte);
		return(-1);
	}
}

int getBMPbit(){
	int value;
	int mask;
	unsigned char nextbit;
	static unsigned char bmpbitsinbuffer = 0;
	static unsigned char bmpbitbuffer = 0;

	value = 0;
	mask = 1;


	if (bmpbitsinbuffer == 0){
		bmpbitbuffer = bmpcurbyte[0];
		bmpcurbyte++;
		bmpbitsinbuffer = 8;
	}
	nextbit = bmpbitbuffer & 0x80;
	if (nextbit) value = 1;
	else value = 0;
	bmpbitbuffer = bmpbitbuffer << 1;
	bmpbitsinbuffer--;

	#ifdef BMP_DEBUG_HIGH
	printf("bitbuffer = %X buffersize = %d nextbit = %d value = %d\n", 
		bmpbitbuffer, bmpbitsinbuffer, nextbit, value);

	#endif
	return(value);
}

int getBMPbuffer(char *buffer, int length){
	memcpy(buffer, bmpcurbyte, length);
	bmpcurbyte += length;
	return(1);
}

void gotoBMPbyte(int byte){
	bmpcurbyte = bmpbuffer + byte;
}

void rewindBMPbytes(int bytes){
	bmpcurbyte -= bytes;
}

void forwardBMPbytes(int bytes){
	bmpcurbyte += bytes;
}



