#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "rgif.h"

//#define GIF_DEBUG
//#define GIF_DEBUG_MED
//#define GIF_DEBUG_HIGH

/* global tables */

unsigned char gifbitsinbuffer = 0;
unsigned char gifbitbuffer = 0;

unsigned char *imagebuffer;
int imagewidth, imageheight;
int imagepos;

int currentoffsetx, currentoffsety;
int currentwidth, currentheight;
int globalmapinfo, globalmappresent, globalbpp, globalcpp;
int localmapinfo, localmappresent, localcpp, localimagetype;

unsigned char *gifcurbyte;
unsigned char *gifbuffer;
unsigned char *gifmaxbyte;
unsigned char background;
int giflength;
long gifbitsread = 0;
int gifbytecounter;
int gifblocksizebyte;
int giflastcode;
int giflastcodenum;

unsigned char g_redcolormap[256];
unsigned char g_greencolormap[256];
unsigned char g_bluecolormap[256];
int g_numcolors;

unsigned char l_redcolormap[256];
unsigned char l_greencolormap[256];
unsigned char l_bluecolormap[256];
int l_numcolors;

int codetable_prefix[4096];
int codetable_suffix[4096];

void initGIFcodetable(int codelength);
int decodeGIFdatablock();
int lookupGIFcode(int num, int code, int *string, int *stringlen);

/*
int main(int argc, char **argv){
	int width;
	int height;
	int comps;
	unsigned char *buffer;
	
	return (readGIF(argv[1], &buffer, &width, &height, &comps));	
}
*/

int readGIF(char *filename, unsigned char **buffer, int *width, int *height, int *components){
	int markertype;
	int endofimage;
	int separator;
	int functioncode;
	int blocklength;
	char imageident[7];

	int i, j, k;
 	gifbitsinbuffer = 0;
	gifbitbuffer = 0;
	imagebuffer = NULL;

	if (openGIFfile(filename) == 0){
		printf("Error opening GIF file %s\n", filename);
		return(0);
	}

	getGIFbuffer(imageident, 6);
	imageident[6] = 0;
	imagewidth = getGIFint();
	imageheight = getGIFint();
	globalmapinfo = getGIFbyte();

	globalmappresent = globalmapinfo >> 7;
	globalbpp = (globalmapinfo & 0x70) >> 4;
	globalcpp = (globalmapinfo & 0x07);
	g_numcolors = 2 << globalcpp;
	background = getGIFbyte();
	getGIFbyte(); 	/* blank */

	#ifdef GIF_DEBUG_LOW
	printf("Image ID: %s\n", imageident); 
	printf("Image Size: (%d,%d)\n", imagewidth, imageheight);
	printf("Global Color Map: %d - BPP %d - CPP %d size %d\n", globalmappresent, globalbpp, globalcpp, g_numcolors);
	printf("Backround: %d\n", background); 
	printf("\n"); 
	#endif

	/* read in global colormap if it exists */
	if (globalmappresent){
		for (i = 0; i < g_numcolors; i++){
			g_redcolormap[i] = getGIFbyte();
			g_greencolormap[i] = getGIFbyte();
			g_bluecolormap[i] = getGIFbyte();
		}
	}

	while (1){
		separator = getGIFbyte();
		if (separator == '!'){
			functioncode = getGIFbyte();
			while(1){
				blocklength = getGIFbyte();
				if (blocklength == 0) break;
				#ifdef GIF_DEBUG_LOW
				printf("Found extension block type %d of length %d\n", functioncode, blocklength);
				#endif
				forwardGIFbytes(blocklength);			
			}
		}
		else if (separator == ','){
			#ifdef GIF_DEBUG_LOW
			printf("Found image descriptor separator %c\n", separator);
			#endif
			currentoffsetx = getGIFint();
			currentoffsety = getGIFint();
			currentwidth = getGIFint();
			currentheight = getGIFint();
			localmapinfo = getGIFbyte();
			localmappresent = localmapinfo >> 7;
			localimagetype = (localmapinfo & 0x40) >> 6;
			localcpp = (localmapinfo & 0x03);

			#ifdef GIF_DEBUG_LOW
			printf("Local Image Descriptor: %d\n", 1); 
			printf("Local Image Position: (%d,%d)\n", currentwidth, currentheight);
			printf("Local Image Size: (%d,%d)\n", currentwidth, currentheight);
			printf("Local Image Type: %d - (0 = Sequential, 1 = Interlaced)\n", localimagetype);
			printf("Local Color Map: %d - CPP %d\n", localmappresent, localcpp);
			#endif

			imagebuffer = calloc(currentwidth * currentheight * 3, 1);
			if (imagebuffer == NULL){
				printf("Couldn't allocate GIF image buffer memory of size %d\n", currentwidth * currentheight * 3);
				return(0);
			}

			/* read in local colormap if it exists */
			l_numcolors = 2 << localcpp;
			if (localmappresent){
				for (i = 0; i < l_numcolors; i++){
					l_redcolormap[i] = getGIFbyte();
					l_greencolormap[i] = getGIFbyte();
					l_bluecolormap[i] = getGIFbyte();
				}
			}
	
			//if (localmappresent) initGIFcodetable(localcpp);
			//else initGIFcodetable(globalcpp);

			imagepos = 0;
			endofimage = 1;
			//while(endofimage){
				endofimage = decodeGIFdatablock();
			//}



			#ifdef GIF_DEBUG_LOW
			printf("%ld/%ld\n", (long)gifcurbyte, (long)gifmaxbyte);
			printf("End of Image %x\n", markertype);
			#endif
			/* if we're done set the imagesize and buffer values */

			*width = imagewidth;
			*height = imageheight;
			*buffer = imagebuffer;
			*components = 3;
			return(1);
		}
	}
}
int decodeGIFdatablock(){
	int mincodesize, codesize;
	int bytes, currpos;
	int code, value;
	int endvalue, resetvalue;
	int string[4096];
	int stringlen;

	int i, j;

	mincodesize = getGIFbyte();
	bytes = getGIFbyte();
	gifblocksizebyte = bytes;
	resetvalue = 1 << mincodesize;
	endvalue = (1 << mincodesize) + 1;

	/* check for end of stream */
	if (bytes == 0){
		#ifdef GIF_DEBUG_LOW
		printf("END OF STREAM\n");
		#endif
		return(0);
	}		

	#ifdef GIF_DEBUG_LOW
	printf("Starting new datablock of size %d with minimum code size %d\n", bytes, mincodesize);
	#endif

	initGIFcodetable(mincodesize);

	currpos = 0;
	codesize = mincodesize;
	gifbytecounter = 0;
	while(1){

		code = getGIFbits(codesize + 1);
		if (code == resetvalue){
			/* reset the tables */
			#ifdef GIF_DEBUG_MED
			printf("Read code %d.. resetting table for codesize %d\n", code, codesize);
			#endif
			initGIFcodetable(mincodesize);
			currpos = 0;
			codesize = mincodesize;
		}
		else if (code == endvalue){
			/* reset the tables */
			#ifdef GIF_DEBUG_LOW
			printf("Read code %d.. ending scan\n", code);
			#endif
			break;
		}
		else {
			lookupGIFcode(currpos, code, string, &stringlen);
			#ifdef GIF_DEBUG_LOW
			printf("code = %d, number of values = %d\n", code, stringlen);
			#endif
			for (j = stringlen - 1; j >= 0; j--){
				#ifdef GIF_DEBUG_MED
				printf("value %d = %d\n", j, string[j]);
				#endif
				if (localmappresent){
					imagebuffer[imagepos++] = l_redcolormap[string[j]];
					imagebuffer[imagepos++] = l_greencolormap[string[j]];
					imagebuffer[imagepos++] = l_bluecolormap[string[j]];
				}
				else{
					imagebuffer[imagepos++] = g_redcolormap[string[j]];
					imagebuffer[imagepos++] = g_greencolormap[string[j]];
					imagebuffer[imagepos++] = g_bluecolormap[string[j]];
				}
			}
			currpos++;

		}
		if (giflastcodenum == (1 << (codesize + 1))){
			/* last code written, reset tables and start from scratch */
			if (giflastcodenum == 4096){
				initGIFcodetable(mincodesize);
				//currpos = 0;
				//#ifdef GIF_DEBUG_MED
				//printf("Resetting codesize from %d to %d\n", codesize, mincodesize);
				//#endif
				//codesize = mincodesize;
			}
			else {
				codesize++;
				#ifdef GIF_DEBUG_MED
				printf("Increasing codesize from %d to %d\n", codesize, codesize + 1);
				#endif
			}
		}
		#ifdef GIF_DEBUG_LOW
		printf("\n");
		#endif
	}
	return(1);
}


int lookupGIFcode(int num, int code, int *value, int *numvalues){
	int prefix;
	int i;
	int first;

	if (num == 0){
		#ifdef GIF_DEBUG_MED
		printf("Fetching first known code %d\n", code);
		#endif
		value[0] = codetable_suffix[code];
		*numvalues = 1;
		giflastcode = code;
	}
	else {
		if (code < giflastcodenum){
			#ifdef GIF_DEBUG_MED
			printf("Fetching known code %d\n", code);
			#endif
			value[0] = codetable_suffix[code];
			prefix = codetable_prefix[code];
			i = 1;
			while (prefix != -1){
				value[i] = codetable_suffix[prefix];
				prefix = codetable_prefix[prefix];
				i++;
			}

			/* find the first code of the lastcode string */
			prefix = codetable_prefix[code];
			first = codetable_suffix[code];
			while (prefix != -1){
				first = codetable_suffix[prefix];
				prefix = codetable_prefix[prefix];
			}
			codetable_prefix[giflastcodenum] = giflastcode;
			codetable_suffix[giflastcodenum] = first;
			//codetable_suffix[giflastcodenum] = codetable_suffix[code];
			#ifdef GIF_DEBUG_MED
			printf("Setting code %d to suffix %d prefix %d\n", giflastcodenum, 
				codetable_suffix[giflastcodenum], codetable_prefix[giflastcodenum]);
			#endif	
			giflastcode = code;
			giflastcodenum++;		
			*numvalues = i;
			return (0);
		}
		else {
			#ifdef GIF_DEBUG_MED
			printf("Fetching unknown code %d\n", code);
			#endif

			value[0] = codetable_suffix[giflastcode];
			value[1] = codetable_suffix[giflastcode];
			prefix = codetable_prefix[giflastcode];
			i = 2;
			while (prefix != -1){
				value[i] = codetable_suffix[prefix];
				prefix = codetable_prefix[prefix];
				i++;
			}
			prefix = codetable_prefix[giflastcode];
			first = codetable_suffix[giflastcode];
			while (prefix != -1){
				first = codetable_suffix[prefix];
				prefix = codetable_prefix[prefix];
			}
			codetable_prefix[giflastcodenum] = giflastcode;
			codetable_suffix[giflastcodenum] = first;

			#ifdef GIF_DEBUG_MED
			printf("Setting code %d to suffix %d prefix %d\n", giflastcodenum, 
				codetable_suffix[giflastcodenum], codetable_prefix[giflastcodenum]);
	
			#endif

			giflastcode = code;
			giflastcodenum++;		
			*numvalues = i;
			return (0);
		}
	}
	return(0);
}

void initGIFcodetable(int codelength){
	int i;
	int rootsize;

	rootsize = 1 << codelength;


	for (i = 0; i < rootsize; i++){
		codetable_prefix[i] = -1;
		codetable_suffix[i] = i;
	} 

	for (i = rootsize; i < 4096; i++){
		codetable_prefix[i] = -1;
		codetable_suffix[i] = -1;
	}
	giflastcodenum = rootsize + 2;
	giflastcode = -1;
}
		
int openGIFfile(char *filename){
	FILE *giffp;
	unsigned char *buffer;
	int size, read;

	giffp = fopen(filename, "rb");
	if (giffp == NULL){
		//printf("Error opening GIF file %s\n", filename);
		return(0);
	}

	size = 0;
	if (fseek(giffp, 0, SEEK_END) == 0){
		size = ftell(giffp);
	}	
	
	if (size == 0) return (0);

	buffer = calloc(size, 1);
	fseek(giffp, 0, SEEK_SET);
	read = fread(buffer, 1, size, giffp);
	if (read != size){
		printf("Error reading GIF file %s, size %d bytes, read %d bytes\n", filename, size, read); 
		return(0);
	}

	gifbuffer = buffer;
	gifcurbyte = buffer; 
	giflength = size;
	gifmaxbyte = gifbuffer + size;

	fclose(giffp);
	return(1);
}

int getGIFint(){
	int value;
	int lsb, msb;

	if (gifcurbyte < gifmaxbyte){ 
		lsb = (int)gifcurbyte[0];
		gifcurbyte++;
		msb = (int)gifcurbyte[0];
		gifcurbyte++;
		value = (msb << 8) | lsb;
		return(value);
	}
	else{
		printf("Fetching out of bounds GIF int %ld/%ld\n", (long)gifcurbyte, (long)gifmaxbyte);
		return(-1); 
	}
}

int getGIFbyte(FILE *jpgfp){
	int value;

	if (gifcurbyte < gifmaxbyte){ 
		value = (int)gifcurbyte[0];
		gifcurbyte++;
		return(value);
	}
	else{
		printf("Fetching out of bounds GIF byte %ld/%ld\n", (long)gifcurbyte, (long)gifmaxbyte);
		return(-1);
	}
}

void rewindGIFbytes(int bytes){
	gifcurbyte -= bytes;
}

void forwardGIFbytes(int bytes){
	gifcurbyte += bytes;
}

/* get a max of 16 bits (sizeof(int)) */
int getGIFbits(int bits){
	int value;
	int mask;
	unsigned char nextbit;

	value = 0;
	mask = 1;
	while (bits){
		if (gifbitsinbuffer == 0){
			if (gifbytecounter == gifblocksizebyte){
				#ifdef GIF_DEBUG_LOW
				printf("Found data block size marker of size %d at %d\n", gifbytecounter, (int)gifcurbyte[0]);
				#endif
				gifcurbyte++;
				gifbytecounter = 0;
			}

			gifbitbuffer = gifcurbyte[0];
			gifcurbyte++;
			gifbytecounter++;
			gifbitsinbuffer = 8;

		}
		nextbit = gifbitbuffer & 0x01;
		if (nextbit) value = value | mask;
		gifbitbuffer = gifbitbuffer >> 1;
		gifbitsinbuffer--;
		bits--;
		mask = mask << 1;

		#ifdef GIF_DEBUG_HIGH
		if (bits == 0) printf("\n");
		printf("bitbuffer = %X buffersize = %d nextbit = %d value = %d bits left %d\n", 
			gifbitbuffer, gifbitsinbuffer, nextbit, value, bits);

		#endif
	}
	return(value);
}

int getGIFbuffer(char *buffer, int length){
	memcpy(buffer, gifcurbyte, length);
	gifcurbyte += length;
	return(1);
}


