#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "rpcx.h"

//#define PCX_DEBUG_LOW
#define PCX_DEBUG_MED
#define PCX_DEBUG_HIGH


unsigned char pcxbitsinbuffer = 0;
unsigned char pcxbitbuffer = 0;

unsigned char *imagebuffer;
int imagewidth, imageheight;
int imagepos;

int currentoffsetx, currentoffsety;
int currentwidth, currentheight;
int globalmapinfo, globalmappresent, globalbpp, globalcpp;
int localmapinfo, localmappresent, localcpp, localimagetype;

unsigned char *pcxcurbyte;
unsigned char *pcxbuffer;
unsigned char *pcxmaxbyte;
unsigned char background;
int pcxlength;
long pcxbitsread = 0;
int pcxbytecounter;
int pcxblocksizebyte;
int pcxlastcode;
int pcxlastcodenum;

unsigned char redcolormap[256];
unsigned char greencolormap[256];
unsigned char bluecolormap[256];
int g_numcolors;

/*
int main(int argc, char **argv){
	int width;
	int height;
	int comps;
	unsigned char *buffer;
	
	return (readPCX(argv[1], &buffer, &width, &height, &comps));	
}
*/

int readPCX(char *filename, unsigned char **buffer, int *width, int *height, int *components){
	unsigned char manufacturer;
	unsigned char version;
	unsigned char encoding;
	unsigned char bpp;
	unsigned char numplanes;
	int startx, starty, endx, endy;
	int resx, resy;
	int bytesperline, bytesperlines;
	int paletteinfo;

	int markertype;
	int endofimage;
	int separator;
	int functioncode;
	int pos, xpos, ypos, lpos, bpos;
	int paletteoffset, code, reps, value, plane;
	int blocklength;
	char imageident[7];

	int i, j, k;
 	pcxbitsinbuffer = 0;
	pcxbitbuffer = 0;
	imagebuffer = NULL;

	if (openPCXfile(filename) == 0){
		printf("Error opening PCX file %s\n", filename);
		return(0);
	}

	manufacturer = getPCXbyte();
	version = getPCXbyte();
	encoding = getPCXbyte();
	bpp = getPCXbyte();
	startx = getPCXint();
	starty = getPCXint();
	endx = getPCXint();
	endy = getPCXint();
	resx = getPCXint();
	resy = getPCXint();
	imagewidth = endx - startx;
	imageheight = endy - starty;
	
	/* get colormap info if it exists */

	/* load the 24 bit palette from the end of the file */
	if (version == 5 && bpp == 8){
		paletteoffset = pcxlength - 769;
		gotoPCXbyte(paletteoffset);
		code = getPCXbyte();
		if (code == 12){
			printf("Found 256 color palette at position %d\n", paletteoffset);
			for (i = 0; i < 256; i++){
				redcolormap[i] = getPCXbyte();
				greencolormap[i] = getPCXbyte();
				bluecolormap[i] = getPCXbyte();
			}
		}
		else {
			printf("Couldn't find 256 color palette at position %d\n", paletteoffset);
		}
	}
					
	gotoPCXbyte(64);

	//for (i = 0; i < 48; i++){
	//	getPCXbyte();
	//}

	getPCXbyte();
	numplanes = getPCXbyte();
	bytesperline = getPCXint();
	bytesperlines = bytesperline * numplanes;
	paletteinfo = getPCXint();

	/* skip to the end of the 128 byte header */
	forwardPCXbytes(58);

	//#ifdef PCX_DEBUG_LOW
	printf("Manufacturer: %d (10 = ZSoft .PCX)\n", manufacturer);
	printf("Version: %d (0 = 2.5, 2 = 2.8 palette, 3 = 2.8 no palette, 5 = 3.0)\n", version);
	printf("Image Position: (%d,%d) - (%d,%d)\n", startx, starty, endx, endy);
	printf("Image Size: (%d,%d)\n", imagewidth, imageheight);
	printf("Device Resolution: %dx%d\n", resx, resy);
	printf("Image BPP: %d\n", bpp);
	printf("Planes: %d\n", numplanes);
	printf("Bytes per line: %d\n", bytesperline);
	printf("Palette info: %d\n", paletteinfo);
	
	printf("\n"); 
	//#endif

	imagebuffer = calloc(imagewidth * imageheight * 3, 1);
	if (imagebuffer == NULL){
		printf("Couldn't allocate PCX image buffer memory of size %d\n", currentwidth * currentheight * 3);
		return(0);
	}

	if (numplanes == 1){
		pos = 0;
		xpos = 0;
		ypos = 0;
		printf("Decoding 256 color PCX\n");
		while(1){
			code = getPCXbyte();
			if ((code & 0xc0) == 0xc0){
				reps = code & 0x3f;
				value = getPCXbyte();

				#ifdef PCX_DEBUG_LOW
				printf("PCX block of size %d value %d\n", reps, value);
				#endif
				for (j = 0; j < reps; j++){
					if (xpos < imagewidth){ 
						imagebuffer[pos] = redcolormap[value];
						pos++;
						imagebuffer[pos] = greencolormap[value];
						pos++;				
						imagebuffer[pos] = bluecolormap[value];
						pos++;
					}
					xpos++;
					if (xpos >= bytesperline){
						ypos++;
						xpos = 0;
					}

				}
			}
			else{
				if (xpos < imagewidth){ 
					imagebuffer[pos] = redcolormap[code];
					pos++;
					imagebuffer[pos] = redcolormap[code];
					pos++;				
					imagebuffer[pos] = redcolormap[code];
					pos++;
				}
				xpos++;

				if (xpos >= bytesperline){
					ypos++;
					xpos = 0;
				}
				#ifdef PCX_DEBUG_LOW
				printf("Single Value %d\n", code);
				#endif
			}
			if (ypos >= imageheight) break;
		}
	}							

	if (numplanes == 3){
		pos = 0;
		xpos = 0;
		ypos = 0;
		lpos = 0;
		plane = 0;
		bpos = 0;

		printf("Decoding 24bit PCX\n");

		while(1){
			code = getPCXbyte();
			if ((code & 0xc0) == 0xc0){
				reps = code & 0x3f;
				value = getPCXbyte();

				#ifdef PCX_DEBUG_LOW
				printf("PCX block of size %d value %d on plane %d at (%d,%d)\n", reps, value, plane, xpos, ypos);
				#endif

				for (j = 0; j < reps; j++){
					if (bpos < bytesperlines && xpos < imagewidth){
						imagebuffer[pos] = value;
						pos += 3;
					}
					bpos++;
					xpos++;
					if (xpos >= imagewidth && plane != 2){ 
						plane++;
						pos = lpos + plane;
						xpos = 0;
					}
				}
			}
			else{
				if (bpos < bytesperlines && xpos < imagewidth){
					imagebuffer[pos] = code;
					pos += 3;
					#ifdef PCX_DEBUG_LOW
					printf("Single Value %d on plane %d at (%d,%d)\n", code, plane, xpos, ypos);
					#endif
				}
				xpos++;
				bpos++;
			}

			if (xpos >= imagewidth && plane != 2){
				plane++;
				pos = lpos + plane;
				xpos = 0;
			}
			if (bpos >= bytesperlines){
				bpos = 0;
				xpos = 0;
				ypos++;
				plane = 0;
				pos = lpos + (3 * imagewidth);
				lpos = pos;
			}

			if (ypos >= imageheight) break;
		}							
	}

	#ifdef GIF_DEBUG_LOW
	printf("%ld/%ld\n", (long)gifcurbyte, (long)gifmaxbyte);
	printf("End of Image %x\n", markertype);
	#endif
	/* if we're done set the imagesize and buffer values */

	*width = imagewidth;
	*height = imageheight;
	*buffer = imagebuffer;
	*components = 3;
	return(1);
}

int openPCXfile(char *filename){
	FILE *pcxfp;
	unsigned char *buffer;
	int size, read;

	pcxfp = fopen(filename, "rb");
	if (pcxfp == NULL){
		//printf("Error opening PCX file %s\n", filename);
		return(0);
	}

	size = 0;
	if (fseek(pcxfp, 0, SEEK_END) == 0){
		size = ftell(pcxfp);
	}	
	
	if (size == 0) return (0);

	buffer = calloc(size, 1);
	fseek(pcxfp, 0, SEEK_SET);
	read = fread(buffer, 1, size, pcxfp);
	if (read != size){
		printf("Error reading PCX file %s, size %d bytes, read %d bytes\n", filename, size, read); 
		return(0);
	}

	pcxbuffer = buffer;
	pcxcurbyte = buffer; 
	pcxlength = size;
	pcxmaxbyte = pcxbuffer + size;

	fclose(pcxfp);
	return(1);
}

int getPCXint(){
	int value;
	int lsb, msb;

	if (pcxcurbyte < pcxmaxbyte){ 
		lsb = (int)pcxcurbyte[0];
		pcxcurbyte++;
		msb = (int)pcxcurbyte[0];
		pcxcurbyte++;
		value = (msb << 8) | lsb;
		return(value);
	}
	else{
		printf("Fetching out of bounds PCX int %ld/%ld\n", (long)pcxcurbyte, (long)pcxmaxbyte);
		return(-1); 
	}
}

int getPCXbyte(FILE *jpgfp){
	int value;

	if (pcxcurbyte < pcxmaxbyte){ 
		value = (int)pcxcurbyte[0];
		pcxcurbyte++;
		return(value);
	}
	else{
		printf("Fetching out of bounds PCX byte %ld/%ld\n", (long)pcxcurbyte, (long)pcxmaxbyte);
		return(-1);
	}
}

void gotoPCXbyte(int byte){
	pcxcurbyte = pcxbuffer + byte;
}

void rewindPCXbytes(int bytes){
	pcxcurbyte -= bytes;
}

void forwardPCXbytes(int bytes){
	pcxcurbyte += bytes;
}

#ifdef LATER
/* get a max of 16 bits (sizeof(int)) */
int getGIFbits(int bits){
	int value;
	int mask;
	unsigned char nextbit;

	value = 0;
	mask = 1;
	while (bits){
		if (gifbitsinbuffer == 0){
			if (gifbytecounter == gifblocksizebyte){
				#ifdef GIF_DEBUG_LOW
				printf("Found data block size marker of size %d at %d\n", gifbytecounter, (int)gifcurbyte[0]);
				#endif
				gifcurbyte++;
				gifbytecounter = 0;
			}

			gifbitbuffer = gifcurbyte[0];
			gifcurbyte++;
			gifbytecounter++;
			gifbitsinbuffer = 8;

		}
		nextbit = gifbitbuffer & 0x01;
		if (nextbit) value = value | mask;
		gifbitbuffer = gifbitbuffer >> 1;
		gifbitsinbuffer--;
		bits--;
		mask = mask << 1;

		#ifdef GIF_DEBUG_HIGH
		if (bits == 0) printf("\n");
		printf("bitbuffer = %X buffersize = %d nextbit = %d value = %d bits left %d\n", 
			gifbitbuffer, gifbitsinbuffer, nextbit, value, bits);

		#endif
	}
	return(value);
}
#endif

int getPCXbuffer(char *buffer, int length){
	memcpy(buffer, pcxcurbyte, length);
	pcxcurbyte += length;
	return(1);
}


